from tests.framework.results.helpers import (
    HARRY_POTTER_CONFIG,
    Hogwarts,
    HogwartsResultsStratifier,
    HousePointsObserver,
    NoStratificationsQuidditchWinsObserver,
    QuidditchWinsObserver,
)



from __future__ import annotations

from collections import defaultdict
from enum import Enum
from typing import TYPE_CHECKING, Callable, Dict, List, Optional, Tuple, Union

import pandas as pd

from vivarium.framework.event import Event
from vivarium.framework.results.context import ResultsContext
from vivarium.framework.values import Pipeline
from vivarium.manager import Manager

if TYPE_CHECKING:
    from vivarium.framework.engine import Builder


class SourceType(Enum):
    COLUMN = 0
    VALUE = 1


class ResultsManager(Manager):
    """Backend manager object for the results management system.

    The :class:`ResultManager` actually performs the actions needed to
    stratify and observe results. It contains the public methods used by the
    :class:`ResultsInterface` to register stratifications and observations,
    which provide it with lists of methods to apply in their respective areas.
    It is able to record observations at any of the time-step sub-steps
    (`time_step__prepare`, `time_step`, `time_step__cleanup`, and
    `collect_metrics`).
    """

    CONFIGURATION_DEFAULTS = {
        "stratification": {
            "default": [],
            "excluded_categories": {},
        }
    }

    def __init__(self) -> None:
        self._raw_results: defaultdict[str, pd.DataFrame] = defaultdict()
        self._results_context = ResultsContext()
        self._required_columns = {"tracked"}
        self._required_values: set[Pipeline] = set()
        self._name = "results_manager"

    @property
    def name(self) -> str:
        """The name of this ResultsManager."""
        return self._name

    def get_results(self) -> Dict[str, pd.DataFrame]:
        """Return the measure-specific formatted results in a dictionary.

        NOTE: self._results_context.observations is a list where each item is a dictionary
        of the form {lifecycle_phase: {(pop_filter, stratifications): List[Observation]}}.
        We use a triple-nested for loop to iterate over only the list of Observations
        (i.e. we do not need the lifecycle_phase, pop_filter, or stratifications).
        """
        formatted = {}
        for observation_details in self._results_context.observations.values():
            for observations in observation_details.values():
                for observation in observations:
                    measure = observation.name
                    results = self._raw_results[measure].copy()
                    formatted[measure] = observation.results_formatter(
                        measure=measure, results=results
                    )
        return formatted

    # noinspection PyAttributeOutsideInit
    def setup(self, builder: "Builder") -> None:
        self._results_context.setup(builder)

        self.logger = builder.logging.get_logger(self.name)
        self.population_view = builder.population.get_view([])
        self.clock = builder.time.clock()
        self.step_size = builder.time.step_size()

        builder.event.register_listener("post_setup", self.on_post_setup)
        builder.event.register_listener("time_step__prepare", self.on_time_step_prepare)
        builder.event.register_listener("time_step", self.on_time_step)
        builder.event.register_listener("time_step__cleanup", self.on_time_step_cleanup)
        builder.event.register_listener("collect_metrics", self.on_collect_metrics)

        self.get_value = builder.value.get_value

        self.set_default_stratifications(builder)

    def on_post_setup(self, _: Event) -> None:
        """Initialize results with 0s DataFrame' for each measure and all stratifications"""
        registered_stratifications = self._results_context.stratifications

        used_stratifications = set()
        for lifecycle_phase in self._results_context.observations:
            for (
                _pop_filter,
                event_requested_stratification_names,
            ), observations in self._results_context.observations[lifecycle_phase].items():
                if event_requested_stratification_names is not None:
                    used_stratifications |= set(event_requested_stratification_names)
                for observation in observations:
                    measure = observation.name
                    self._raw_results[measure] = observation.results_initializer(
                        event_requested_stratification_names, registered_stratifications
                    )

        registered_stratification_names = set(
            stratification.name for stratification in registered_stratifications
        )
        unused_stratifications = registered_stratification_names - used_stratifications
        if unused_stratifications:
            self.logger.info(
                "The following stratifications are registered but not used by any "
                f"observers: \n{sorted(list(unused_stratifications))}"
            )
        missing_stratifications = used_stratifications - registered_stratification_names
        if missing_stratifications:
            raise ValueError(
                "The following observers are requested to be stratified by "
                f"stratifications that are not registered: \n{sorted(list(missing_stratifications))}"
            )

    def on_time_step_prepare(self, event: Event) -> None:
        self.gather_results("time_step__prepare", event)

    def on_time_step(self, event: Event) -> None:
        self.gather_results("time_step", event)

    def on_time_step_cleanup(self, event: Event) -> None:
        self.gather_results("time_step__cleanup", event)

    def on_collect_metrics(self, event: Event) -> None:
        self.gather_results("collect_metrics", event)

    def gather_results(self, lifecycle_phase: str, event: Event) -> None:
        """Update the existing results with new results. Any columns in the
        results group that are not already in the existing results are initialized
        with 0.0.
        """
        population = self._prepare_population(event)
        if population.empty:
            return
        for results_group, measure, updater in self._results_context.gather_results(
            population, lifecycle_phase, event
        ):
            if results_group is not None and measure is not None and updater is not None:
                self._raw_results[measure] = updater(
                    self._raw_results[measure], results_group
                )

    ##########################
    # Stratification methods #
    ##########################

    def set_default_stratifications(self, builder: Builder) -> None:
        default_stratifications = builder.configuration.stratification.default
        self._results_context.set_default_stratifications(default_stratifications)

    def register_stratification(
        self,
        name: str,
        categories: List[str],
        excluded_categories: Optional[List[str]],
        mapper: Optional[Callable[[Union[pd.Series[str], pd.DataFrame]], pd.Series[str]]],
        is_vectorized: bool,
        requires_columns: List[str] = [],
        requires_values: List[str] = [],
    ) -> None:
        """Manager-level stratification registration, including resources and the stratification itself.

        Parameters
        ----------
        name
            Name of the of the column created by the stratification.
        categories
            List of string values that the mapper is allowed to output.
        excluded_categories
            List of mapped string values to be excluded from results processing.
            If None (the default), will use exclusions as defined in the configuration.
        mapper
            A callable that emits values in `categories` given inputs from columns
            and values in the `requires_columns` and `requires_values`, respectively.
        is_vectorized
            `True` if the mapper function expects a `DataFrame`, and `False` if it
            expects a row of the `DataFrame` and should be used by calling :func:`df.apply`.
        requires_columns
            A list of the state table columns that already need to be present
            and populated in the state table before the pipeline modifier
            is called.
        requires_values
            A list of the value pipelines that need to be properly sourced
            before the pipeline modifier is called.

        Returns
        ------
        None
        """
        self.logger.debug(f"Registering stratification {name}")
        target_columns = list(requires_columns) + list(requires_values)
        self._results_context.add_stratification(
            name, target_columns, categories, excluded_categories, mapper, is_vectorized
        )
        self._add_resources(requires_columns, SourceType.COLUMN)
        self._add_resources(requires_values, SourceType.VALUE)

    def register_binned_stratification(
        self,
        target: str,
        binned_column: str,
        bin_edges: List[Union[int, float]],
        labels: List[str],
        excluded_categories: Optional[List[str]],
        target_type: str,
        **cut_kwargs,
    ) -> None:
        """Manager-level registration of a continuous `target` quantity to observe into bins in a `binned_column`.

        Parameters
        ----------
        target
            String name of the state table column or value pipeline used to stratify.
        binned_column
            String name of the column for the binned quantities.
        bin_edges
            List of scalars defining the bin edges, passed to :meth: pandas.cut.
            The length must equal the length of `labels` plus one.
            Note that the bins are left edge inclusive, e.g. bin edges [1, 2, 3]
            indicate groups [1, 2) and [2, 3).
        labels
            List of string labels for bins. The length must equal to the length
            of `bin_edges` minus one.
        excluded_categories
            List of mapped string values to be excluded from results processing.
            If None (the default), will use exclusions as defined in the configuration.
        target_type
            "column" or "value"
        **cut_kwargs
            Keyword arguments for :meth: pandas.cut.

        Returns
        ------
        None
        """

        def _bin_data(data: Union[pd.Series, pd.DataFrame]) -> pd.Series:
            data = data.squeeze()
            if not isinstance(data, pd.Series):
                raise ValueError(f"Expected a Series, but got type {type(data)}.")
            return pd.cut(
                data, bin_edges, labels=labels, right=False, include_lowest=True, **cut_kwargs
            )

        if len(bin_edges) != len(labels) + 1:
            raise ValueError(
                f"The number of bin edges plus 1 ({len(bin_edges)+1}) does not "
                f"match the number of labels ({len(labels)})"
            )

        target_arg = "requires_columns" if target_type == "column" else "required_values"
        target_kwargs = {target_arg: [target]}

        self.register_stratification(
            name=binned_column,
            categories=labels,
            excluded_categories=excluded_categories,
            mapper=_bin_data,
            is_vectorized=True,
            **target_kwargs,
        )

    def register_observation(
        self,
        observation_type,
        is_stratified: bool,
        name: str,
        pop_filter: str,
        when: str,
        requires_columns: List[str],
        requires_values: List[str],
        **kwargs,
    ):
        self.logger.debug(f"Registering observation {name}")

        if is_stratified:
            additional_stratifications = kwargs.get("additional_stratifications", [])
            excluded_stratifications = kwargs.get("excluded_stratifications", [])
            self._warn_check_stratifications(
                additional_stratifications, excluded_stratifications
            )
            stratifications = self._get_stratifications(
                kwargs.get("stratifications", []),
                additional_stratifications,
                excluded_stratifications,
            )
            kwargs["stratifications"] = stratifications
            del kwargs["additional_stratifications"]
            del kwargs["excluded_stratifications"]

        self._add_resources(requires_columns, SourceType.COLUMN)
        self._add_resources(requires_values, SourceType.VALUE)

        self._results_context.register_observation(
            observation_type=observation_type,
            name=name,
            pop_filter=pop_filter,
            when=when,
            **kwargs,
        )

    ##################
    # Helper methods #
    ##################

    def _get_stratifications(
        self,
        stratifications: List[str] = [],
        additional_stratifications: List[str] = [],
        excluded_stratifications: List[str] = [],
    ) -> Tuple[str, ...]:
        stratifications = list(
            set(
                self._results_context.default_stratifications
                + stratifications
                + additional_stratifications
            )
            - set(excluded_stratifications)
        )
        # Makes sure measure identifiers have fields in the same relative order.
        return tuple(sorted(stratifications))

    def _add_resources(self, target: List[str], target_type: SourceType) -> None:
        if len(target) == 0:
            return  # do nothing on empty lists
        target_set = set(target) - {"event_time", "current_time", "event_step_size"}
        if target_type == SourceType.COLUMN:
            self._required_columns.update(target_set)
        elif target_type == SourceType.VALUE:
            self._required_values.update([self.get_value(target) for target in target_set])

    def _prepare_population(self, event: Event) -> pd.DataFrame:
        population = self.population_view.subview(list(self._required_columns)).get(
            event.index
        )
        population["current_time"] = self.clock()
        population["event_step_size"] = event.step_size
        population["event_time"] = self.clock() + event.step_size
        for k, v in event.user_data.items():
            population[k] = v
        for pipeline in self._required_values:
            population[pipeline.name] = pipeline(event.index)
        return population

    def _warn_check_stratifications(
        self, additional_stratifications: List[str], excluded_stratifications: List[str]
    ) -> None:
        """Check additional and excluded stratifications if they'd not affect
        stratifications (i.e., would be NOP), and emit warning."""
        nop_additional = [
            s
            for s in additional_stratifications
            if s in self._results_context.default_stratifications
        ]
        if len(nop_additional):
            self.logger.warning(
                f"Specified additional stratifications are already included by default: {nop_additional}",
            )
        nop_exclude = [
            s
            for s in excluded_stratifications
            if s not in self._results_context.default_stratifications
        ]
        if len(nop_exclude):
            self.logger.warning(
                f"Specified excluded stratifications are already not included by default: {nop_exclude}",
            )



from typing import Any, Dict, List, Optional

import pandas as pd

from vivarium.framework.engine import Builder
from vivarium.framework.results import Observer as Observer_


class Observer(Observer_):
    ##############
    # Properties #
    ##############

    @property
    def configuration_defaults(self) -> Dict[str, Any]:
        return {
            "mortality": {
                "life_expectancy": 80,
            }
        }

    @property
    def columns_required(self) -> Optional[List[str]]:
        return ["age", "alive"]

    def register_observations(self, builder: Builder) -> None:
        builder.results.register_adding_observation(
            name="total_population_alive",
            requires_columns=["alive"],
            pop_filter='alive == "alive"',
        )
        builder.results.register_adding_observation(
            name="total_population_dead",
            requires_columns=["alive"],
            pop_filter='alive == "dead"',
        )
        builder.results.register_adding_observation(
            name="years_of_life_lost",
            requires_columns=["age", "alive"],
            aggregator=self.calculate_ylls,
        )

    def calculate_ylls(self, df: pd.DataFrame) -> float:
        return (self.life_expectancy - df.loc[df["alive"] == "dead", "age"]).sum()

    #####################
    # Lifecycle methods #
    #####################

    # noinspection PyAttributeOutsideInit
    def setup(self, builder: Builder) -> None:
        self.life_expectancy = builder.configuration.mortality.life_expectancy



from __future__ import annotations

from dataclasses import dataclass
from typing import Callable, List, Optional, Union

import pandas as pd
from pandas.api.types import CategoricalDtype

STRATIFICATION_COLUMN_SUFFIX: str = "mapped_values"


@dataclass
class Stratification:
    """Class for stratifying observed quantities by specified characteristics

    Each Stratification represents a set of mutually exclusive and collectively
    exhaustive categories into which simulants can be assigned.

    The `Stratification` class has six fields: `name`, `sources`, `mapper`,
    `categories`, `excluded_categories`, and `is_vectorized`. The `name` is the
    name of the column created by the mapper. The `sources` is a list of columns
    in the extended state table that are the inputs to the mapper function. Simulants
    will later be grouped by this column (or these columns) during stratification.
    `categories` is the total set of values that the mapper can output.
    `excluded_categories` are values that have been requested to be excluded (and
    already removed) from `categories`. The `mapper` is the method that transforms the source
    to the name column. The method produces an output column by calling the mapper on the source
    columns. If the mapper is `None`, the default identity mapper is used. If
    the mapper is not vectorized this is performed by using `pd.apply`.
    Finally, `is_vectorized` is a boolean parameter that signifies whether
    mapper function is applied to a single simulant (`False`) or to the whole
    population (`True`).

    `Stratification` also has a `__call__()` method. The method produces an
    output column by calling the mapper on the source columns.
    """

    name: str
    sources: List[str]
    categories: List[str]
    excluded_categories: List[str]
    mapper: Optional[Callable[[Union[pd.Series[str], pd.DataFrame]], pd.Series[str]]] = None
    is_vectorized: bool = False

    def __str__(self) -> str:
        return (
            f"Stratification '{self.name}' with sources {self.sources}, "
            f"categories {self.categories}, and mapper {self.mapper.__name__}"
        )

    def __post_init__(self) -> None:
        if self.mapper is None:
            if len(self.sources) != 1:
                raise ValueError(
                    f"No mapper provided for stratification {self.name} with "
                    f"{len(self.sources)} stratification sources."
                )
            self.mapper = self._default_mapper
            self.is_vectorized = True
        if not self.categories:
            raise ValueError("The categories argument must be non-empty.")
        if not self.sources:
            raise ValueError("The sources argument must be non-empty.")

    def __call__(self, population: pd.DataFrame) -> pd.Series[str]:
        """Apply the mapper to the population 'sources' columns and add the result
        to the population. Any excluded categories (which have already been removed
        from self.categories) will be converted to NaNs in the new column
        and dropped later at the observation level.
        """
        if self.is_vectorized:
            mapped_column = self.mapper(population[self.sources])
        else:
            mapped_column = population[self.sources].apply(self.mapper, axis=1)
        unknown_categories = set(mapped_column) - set(
            self.categories + self.excluded_categories
        )
        # Reduce all nans to a single one
        unknown_categories = [cat for cat in unknown_categories if not pd.isna(cat)]
        if mapped_column.isna().any():
            unknown_categories.append(mapped_column[mapped_column.isna()].iat[0])
        if unknown_categories:
            raise ValueError(f"Invalid values mapped to {self.name}: {unknown_categories}")

        # Convert the dtype to the allowed categories. Note that this will
        # result in Nans for any values in excluded_categories.
        mapped_column = mapped_column.astype(
            CategoricalDtype(categories=self.categories, ordered=True)
        )
        return mapped_column

    @staticmethod
    def _default_mapper(pop: pd.DataFrame) -> pd.Series[str]:
        return pop.squeeze(axis=1)


def get_mapped_col_name(col_name: str) -> str:
    """Return a new column name to be used for mapped values"""
    return f"{col_name}_{STRATIFICATION_COLUMN_SUFFIX}"


def get_original_col_name(col_name: str) -> str:
    """Return the original column name given a modified mapped column name."""
    return (
        col_name[: -(len(STRATIFICATION_COLUMN_SUFFIX)) - 1]
        if col_name.endswith(f"_{STRATIFICATION_COLUMN_SUFFIX}")
        else col_name
    )



from abc import ABC, abstractmethod
from typing import Any, Dict

from vivarium import Component
from vivarium.framework.engine import Builder


class Observer(Component, ABC):
    """An abstract base class intended to be subclassed by observer components.
    The primary purpose of this class is to provide attributes required by
    the subclass `report` method.

    Note that a `register_observation` method must be defined in the subclass.
    """

    def __init__(self) -> None:
        super().__init__()
        self.results_dir = None

    @property
    def configuration_defaults(self) -> Dict[str, Any]:
        return {
            "stratification": {
                self.get_configuration_name(): {
                    "exclude": [],
                    "include": [],
                },
            },
        }

    def get_configuration_name(self) -> str:
        return self.name.split("_observer")[0]

    @abstractmethod
    def register_observations(self, builder: Builder) -> None:
        """(Required). Register observations with within each observer."""
        pass

    def setup_component(self, builder: Builder) -> None:
        super().setup_component(builder)
        self.register_observations(builder)
        self.get_formatter_attributes(builder)

    def get_formatter_attributes(self, builder: Builder) -> None:
        """Define commonly-used attributes for reporting."""
        self.results_dir = (
            builder.configuration.to_dict()
            .get("output_data", {})
            .get("results_directory", None)
        )



from __future__ import annotations

from collections import defaultdict
from typing import Callable, Generator, List, Optional, Tuple, Type, Union

import pandas as pd
from pandas.core.groupby import DataFrameGroupBy

from vivarium.framework.engine import Builder
from vivarium.framework.event import Event
from vivarium.framework.results.exceptions import ResultsConfigurationError
from vivarium.framework.results.observation import BaseObservation
from vivarium.framework.results.stratification import (
    Stratification,
    get_mapped_col_name,
    get_original_col_name,
)


class ResultsContext:
    """
    Manager context for organizing observations and the stratifications they require.

    This context object is wholly contained by the manager :class:`vivarium.framework.results.manager.ResultsManger`.
    Stratifications can be added to the context through the manager via the
    :meth:`vivarium.framework.results.context.ResultsContext.add_observation` method.
    """

    def __init__(self) -> None:
        self.default_stratifications: List[str] = []
        self.stratifications: List[Stratification] = []
        self.excluded_categories: dict[str, list[str]] = {}
        # keys are event names: [
        #     "time_step__prepare",
        #     "time_step",
        #     "time_step__cleanup",
        #     "collect_metrics",
        # ]
        # values are dicts with
        #     key (filter, grouper)
        #     value Observation
        self.observations: defaultdict = defaultdict(lambda: defaultdict(list))

    @property
    def name(self) -> str:
        return "results_context"

    def setup(self, builder: Builder) -> None:
        self.logger = builder.logging.get_logger(self.name)
        self.excluded_categories = (
            builder.configuration.stratification.excluded_categories.to_dict()
        )

    # noinspection PyAttributeOutsideInit
    def set_default_stratifications(self, default_grouping_columns: List[str]) -> None:
        if self.default_stratifications:
            raise ResultsConfigurationError(
                "Multiple calls are being made to set default grouping columns "
                "for results production."
            )
        self.default_stratifications = default_grouping_columns

    def add_stratification(
        self,
        name: str,
        sources: List[str],
        categories: List[str],
        excluded_categories: Optional[List[str]],
        mapper: Optional[Callable[[Union[pd.Series[str], pd.DataFrame]], pd.Series[str]]],
        is_vectorized: bool,
    ) -> None:
        """Add a stratification to the context.

        Parameters
        ----------
        name
            Name of the of the column created by the stratification.
        sources
            A list of the columns and values needed for the mapper to determinate
            categorization.
        categories
            List of string values that the mapper is allowed to output.
        excluded_categories
            List of mapped string values to be excluded from results processing.
            If None (the default), will use exclusions as defined in the configuration.
        mapper
            A callable that emits values in `categories` given inputs from columns
            and values in the `requires_columns` and `requires_values`, respectively.
        is_vectorized
            `True` if the mapper function expects a `DataFrame`, and `False` if it
            expects a row of the `DataFrame` and should be used by calling :func:`df.apply`.


        Returns
        ------
        None

        """
        already_used = [
            stratification
            for stratification in self.stratifications
            if stratification.name == name
        ]
        if already_used:
            raise ValueError(
                f"Stratification name '{name}' is already used: {str(already_used[0])}."
            )
        unique_categories = set(categories)
        if len(categories) != len(unique_categories):
            for category in unique_categories:
                categories.remove(category)
            raise ValueError(
                f"Found duplicate categories in stratification '{name}': {categories}."
            )

        # Handle excluded categories. If excluded_categories are explicitly
        # passed in, we use that instead of what is in the model spec.
        to_exclude = (
            excluded_categories
            if excluded_categories is not None
            else self.excluded_categories.get(name, [])
        )
        unknown_exclusions = set(to_exclude) - set(categories)
        if len(unknown_exclusions) > 0:
            raise ValueError(
                f"Excluded categories {unknown_exclusions} not found in categories "
                f"{categories} for stratification '{name}'."
            )
        if to_exclude:
            self.logger.debug(
                f"'{name}' has category exclusion requests: {to_exclude}\n"
                "Removing these from the allowable categories."
            )
            categories = [category for category in categories if category not in to_exclude]

        stratification = Stratification(
            name,
            sources,
            categories,
            to_exclude,
            mapper,
            is_vectorized,
        )
        self.stratifications.append(stratification)

    def register_observation(
        self,
        observation_type: Type[BaseObservation],
        name: str,
        pop_filter: str,
        when: str,
        **kwargs,
    ) -> None:
        """Add an observation to the context.

        Parameters
        ----------
        observation_type
            Class type of the observation to register.
        name
            Name of the metric to observe and result file.
        pop_filter
            A Pandas query filter string to filter the population down to the
            simulants who should be considered for the observation.
        when
            String name of the phase of a time-step the observation should happen.
            Valid values are: `"time_step__prepare"`, `"time_step"`,
            `"time_step__cleanup"`, `"collect_metrics"`.
        kwargs
            Additional keyword arguments to pass to the observation constructor.


        Returns
        ------
        None

        """
        already_used = None
        if self.observations:
            # NOTE: self.observations is a list where each item is a dictionary
            # of the form {lifecycle_phase: {(pop_filter, stratifications): List[Observation]}}.
            # We use a triple-nested for loop to iterate over only the list of Observations
            # (i.e. we do not need the lifecycle_phase, pop_filter, or stratifications).
            for observation_details in self.observations.values():
                for observations in observation_details.values():
                    for observation in observations:
                        if observation.name == name:
                            already_used = observation
        if already_used:
            raise ValueError(
                f"Observation name '{name}' is already used: {str(already_used)}."
            )
        observation = observation_type(name=name, pop_filter=pop_filter, when=when, **kwargs)
        self.observations[observation.when][
            (observation.pop_filter, observation.stratifications)
        ].append(observation)

    def gather_results(
        self, population: pd.DataFrame, lifecycle_phase: str, event: Event
    ) -> Generator[
        Tuple[
            Optional[pd.DataFrame],
            Optional[str],
            Optional[Callable[[pd.DataFrame, pd.DataFrame], pd.DataFrame]],
        ],
        None,
        None,
    ]:
        """Generate current results for all observations at this lifecycle phase and event."""

        for stratification in self.stratifications:
            # Add new columns of mapped values to the population to prevent name collisions
            new_column = get_mapped_col_name(stratification.name)
            if new_column in population.columns:
                raise ValueError(
                    f"Stratification column '{new_column}' "
                    "already exists in the state table or as a pipeline which is a required "
                    "name for stratifying results - choose a different name."
                )
            population[new_column] = stratification(population)

        # Optimization: We store all the producers by pop_filter and stratifications
        # so that we only have to apply them once each time we compute results.
        for (pop_filter, stratification_names), observations in self.observations[
            lifecycle_phase
        ].items():
            # Results production can be simplified to
            # filter -> groupby -> aggregate in all situations we've seen.
            filtered_pop = self._filter_population(
                population, pop_filter, stratification_names
            )
            if filtered_pop.empty:
                yield None, None, None
            else:
                if stratification_names is None:
                    pop = filtered_pop
                else:
                    pop = self._get_groups(stratification_names, filtered_pop)
                for observation in observations:
                    results = observation.observe(event, pop, stratification_names)
                    if results is not None:
                        self._rename_stratification_columns(results)

                    yield (results, observation.name, observation.results_updater)

    def _filter_population(
        self,
        population: pd.DataFrame,
        pop_filter: str,
        stratification_names: Optional[tuple[str, ...]],
    ) -> pd.DataFrame:
        """Filter the population based on the filter string as well as any
        excluded stratification categories
        """
        pop = population.query(pop_filter) if pop_filter else population.copy()
        if stratification_names:
            # Drop all rows in the mapped_stratification columns that have NaN values
            # (which only exist if the mapper returned an excluded category).
            pop = pop.dropna(
                subset=[
                    get_mapped_col_name(stratification)
                    for stratification in stratification_names
                ]
            )
        return pop

    @staticmethod
    def _get_groups(
        stratifications: Tuple[str, ...], filtered_pop: pd.DataFrame
    ) -> DataFrameGroupBy:
        """Group the population by stratifications.
        NOTE: Stratifications at this point can be an empty tuple.
        HACK: It's a bit hacky how we are handling the groupby object if there
        are no stratifications. The alternative is to use the entire population
        instead of a groupby object, but then we would need to handle
        the different ways the aggregator can behave.
        """

        if stratifications:
            pop_groups = filtered_pop.groupby(
                [get_mapped_col_name(stratification) for stratification in stratifications],
                observed=False,
            )
        else:
            pop_groups = filtered_pop.groupby(lambda _: "all")
        return pop_groups

    def _rename_stratification_columns(self, results: pd.DataFrame) -> None:
        """convert stratified mapped index names to original"""
        if isinstance(results.index, pd.MultiIndex):
            idx_names = [get_original_col_name(name) for name in results.index.names]
            results.rename_axis(index=idx_names, inplace=True)
        else:
            idx_name = results.index.name
            if idx_name is not None:
                results.index.rename(get_original_col_name(idx_name), inplace=True)



from __future__ import annotations

import itertools
from abc import ABC
from dataclasses import dataclass
from typing import Callable, Optional, Tuple, Union

import pandas as pd
from pandas.api.types import CategoricalDtype
from pandas.core.groupby import DataFrameGroupBy

from vivarium.framework.event import Event
from vivarium.framework.results.stratification import Stratification

VALUE_COLUMN = "value"


@dataclass
class BaseObservation(ABC):
    """An abstract base dataclass to be inherited by concrete observations.
    This class includes the following attributes:
    - `name`: name of the observation and is the measure it is observing
    - `pop_filter`: a filter that is applied to the population before the observation is made
    - `when`: the phase that the observation is registered to
    - `results_initializer`: method that initializes the results
    - `results_gatherer`: method that gathers the new observation results
    - `results_updater`: method that updates the results with new observations
    - `results_formatter`: method that formats the results
    - `to_observe`: method that determines whether to observe an event
    """

    name: str
    pop_filter: str
    when: str
    results_initializer: Callable[..., pd.DataFrame]
    results_gatherer: Callable[..., pd.DataFrame]
    results_updater: Callable[[pd.DataFrame, pd.DataFrame], pd.DataFrame]
    results_formatter: Callable[[str, pd.DataFrame], pd.DataFrame]
    stratifications: Optional[Tuple[str]]
    to_observe: Callable[[Event], bool]

    def observe(
        self,
        event: Event,
        df: Union[pd.DataFrame, DataFrameGroupBy],
        stratifications: Optional[tuple[str, ...]],
    ) -> Optional[pd.DataFrame]:
        if not self.to_observe(event):
            return None
        else:
            if stratifications is None:
                return self.results_gatherer(df)
            else:
                return self.results_gatherer(df, stratifications)


class UnstratifiedObservation(BaseObservation):
    """Container class for managing unstratified observations.
    This class includes the following attributes:
    - `name`: name of the observation and is the measure it is observing
    - `pop_filter`: a filter that is applied to the population before the observation is made
    - `when`: the phase that the observation is registered to
    - `results_gatherer`: method that gathers the new observation results
    - `results_updater`: method that updates the results with new observations
    - `results_formatter`: method that formats the results
    - `to_observe`: method that determines whether to observe an event
    """

    def __init__(
        self,
        name: str,
        pop_filter: str,
        when: str,
        results_gatherer: Callable[[pd.DataFrame], pd.DataFrame],
        results_updater: Callable[[pd.DataFrame, pd.DataFrame], pd.DataFrame],
        results_formatter: Callable[[str, pd.DataFrame], pd.DataFrame],
        to_observe: Callable[[Event], bool] = lambda event: True,
    ):
        super().__init__(
            name=name,
            pop_filter=pop_filter,
            when=when,
            results_initializer=self.initialize_results,
            results_gatherer=results_gatherer,
            results_updater=results_updater,
            results_formatter=results_formatter,
            stratifications=None,
            to_observe=to_observe,
        )

    @staticmethod
    def initialize_results(
        requested_stratification_names: set[str],
        registered_stratifications: list[Stratification],
    ) -> pd.DataFrame:
        """Initialize an empty dataframe."""
        return pd.DataFrame()


class StratifiedObservation(BaseObservation):
    """Container class for managing stratified observations.
    This class includes the following attributes:
    - `name`: name of the observation and is the measure it is observing
    - `pop_filter`: a filter that is applied to the population before the observation is made
    - `when`: the phase that the observation is registered to
    - `results_updater`: method that updates the results with new observations
    - `results_formatter`: method that formats the results
    - `stratifications`: a tuple of columns for the observation to stratify by
    - `aggregator_sources`: a list of the columns to observe
    - `aggregator`: a method that aggregates the `aggregator_sources`
    - `to_observe`: method that determines whether to observe an event
    """

    def __init__(
        self,
        name: str,
        pop_filter: str,
        when: str,
        results_updater: Callable[[pd.DataFrame, pd.DataFrame], pd.DataFrame],
        results_formatter: Callable[[str, pd.DataFrame], pd.DataFrame],
        stratifications: Tuple[str, ...],
        aggregator_sources: Optional[list[str]],
        aggregator: Callable[[pd.DataFrame], Union[float, pd.Series[float]]],
        to_observe: Callable[[Event], bool] = lambda event: True,
    ):
        super().__init__(
            name=name,
            pop_filter=pop_filter,
            when=when,
            results_initializer=self.initialize_results,
            results_gatherer=self.results_gatherer,
            results_updater=results_updater,
            results_formatter=results_formatter,
            stratifications=stratifications,
            to_observe=to_observe,
        )
        self.aggregator_sources = aggregator_sources
        self.aggregator = aggregator

    @staticmethod
    def initialize_results(
        requested_stratification_names: set[str],
        registered_stratifications: list[Stratification],
    ) -> pd.DataFrame:
        """Initialize a dataframe of 0s with complete set of stratifications as the index."""

        # Set up the complete index of all used stratifications
        requested_and_registered_stratifications = [
            stratification
            for stratification in registered_stratifications
            if stratification.name in requested_stratification_names
        ]
        stratification_values = {
            stratification.name: stratification.categories
            for stratification in requested_and_registered_stratifications
        }
        if stratification_values:
            stratification_names = list(stratification_values.keys())
            df = pd.DataFrame(
                list(itertools.product(*stratification_values.values())),
                columns=stratification_names,
            ).astype(CategoricalDtype())
        else:
            # We are aggregating the entire population so create a single-row index
            stratification_names = ["stratification"]
            df = pd.DataFrame(["all"], columns=stratification_names).astype(
                CategoricalDtype()
            )

        # Initialize a zeros dataframe
        df[VALUE_COLUMN] = 0.0
        df = df.set_index(stratification_names)

        return df

    def results_gatherer(
        self,
        pop_groups: DataFrameGroupBy,
        stratifications: Tuple[str, ...],
    ) -> pd.DataFrame:
        df = self._aggregate(pop_groups, self.aggregator_sources, self.aggregator)
        df = self._format(df)
        df = self._expand_index(df)
        if not list(stratifications):
            df.index.name = "stratification"
        return df

    @staticmethod
    def _aggregate(
        pop_groups: DataFrameGroupBy,
        aggregator_sources: Optional[list[str]],
        aggregator: Callable[[pd.DataFrame], Union[float, pd.Series[float]]],
    ) -> Union[pd.Series[float], pd.DataFrame]:
        aggregates = (
            pop_groups[aggregator_sources].apply(aggregator).fillna(0.0)
            if aggregator_sources
            else pop_groups.apply(aggregator)
        ).astype(float)
        return aggregates

    @staticmethod
    def _format(aggregates: Union[pd.Series[float], pd.DataFrame]) -> pd.DataFrame:
        df = pd.DataFrame(aggregates) if isinstance(aggregates, pd.Series) else aggregates
        if df.shape[1] == 1:
            df.rename(columns={df.columns[0]: "value"}, inplace=True)
        return df

    @staticmethod
    def _expand_index(aggregates: pd.DataFrame) -> pd.DataFrame:
        if isinstance(aggregates.index, pd.MultiIndex):
            full_idx = pd.MultiIndex.from_product(aggregates.index.levels)
        else:
            full_idx = aggregates.index
        aggregates = aggregates.reindex(full_idx).fillna(0.0)
        return aggregates


class AddingObservation(StratifiedObservation):
    """Specific container class for managing stratified observations that add
    new results to previous ones at each phase the class is registered to.
    This class includes the following attributes:
    - `name`: name of the observation and is the measure it is observing
    - `pop_filter`: a filter that is applied to the population before the observation is made
    - `when`: the phase that the observation is registered to
    - `results_formatter`: method that formats the results
    - `stratifications`: a tuple of columns for the observation to stratify by
    - `aggregator_sources`: a list of the columns to observe
    - `aggregator`: a method that aggregates the `aggregator_sources`
    - `to_observe`: method that determines whether to observe an event
    """

    def __init__(
        self,
        name: str,
        pop_filter: str,
        when: str,
        results_formatter: Callable[[str, pd.DataFrame], pd.DataFrame],
        stratifications: Tuple[str, ...],
        aggregator_sources: Optional[list[str]],
        aggregator: Callable[[pd.DataFrame], Union[float, pd.Series[float]]],
        to_observe: Callable[[Event], bool] = lambda event: True,
    ):
        super().__init__(
            name=name,
            pop_filter=pop_filter,
            when=when,
            results_updater=self.add_results,
            results_formatter=results_formatter,
            stratifications=stratifications,
            aggregator_sources=aggregator_sources,
            aggregator=aggregator,
            to_observe=to_observe,
        )

    @staticmethod
    def add_results(
        existing_results: pd.DataFrame, new_observations: pd.DataFrame
    ) -> pd.DataFrame:
        updated_results = existing_results.copy()
        # Look for extra columns in the new_observations and initialize with 0.
        extra_cols = [
            c for c in new_observations.columns if c not in existing_results.columns
        ]
        if extra_cols:
            updated_results[extra_cols] = 0.0
        for col in new_observations.columns:
            updated_results[col] += new_observations[col]
        return updated_results


class ConcatenatingObservation(UnstratifiedObservation):
    """Specific container class for managing observations that concatenate
    new results to previous ones at each phase the class is registered to.
    Note that this class does not support stratifications.
    This class includes the following attributes:
    - `name`: name of the observation and is the measure it is observing
    - `pop_filter`: a filter that is applied to the population before the observation is made
    - `when`: the phase that the observation is registered to
    - `included_columns`: the columns to include in the observation
    - `results_formatter`: method that formats the results
    - `to_observe`: method that determines whether to observe an event
    """

    def __init__(
        self,
        name: str,
        pop_filter: str,
        when: str,
        included_columns: list[str],
        results_formatter: Callable[[str, pd.DataFrame], pd.DataFrame],
        to_observe: Callable[[Event], bool] = lambda event: True,
    ):
        super().__init__(
            name=name,
            pop_filter=pop_filter,
            when=when,
            results_gatherer=self.results_gatherer,
            results_updater=self.concatenate_results,
            results_formatter=results_formatter,
            to_observe=to_observe,
        )
        self.included_columns = included_columns

    @staticmethod
    def concatenate_results(
        existing_results: pd.DataFrame, new_observations: pd.DataFrame
    ) -> pd.DataFrame:
        if existing_results.empty:
            return new_observations
        return pd.concat([existing_results, new_observations], axis=0).reset_index(drop=True)

    def results_gatherer(self, pop: pd.DataFrame) -> pd.DataFrame:
        return pop[self.included_columns]
